#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/17 17:21
# @email    : 1747193328@qq.com
"""Aggregate top-level UI pages for the NEP Toolkit application."""
from .makedata import MakeDataWidget
from .settings import SettingsWidget
from .show_nep import ShowNepWidget
from .data_manager import DataManagerWidget
